#!/bin/sh
#Example use of distributed optimised online genetic algorithm with niche mechanism.
#To make use of this file:
# - log in https://oga.jod.li
# - change the username and key below (u and k) with the values on the web page
# - optionnally adapt the population parameters
# - place an executable called getfitness within the same directory that reads a chromosome from stdin and writes the fitness in stdout
# - execute this script
#More information available here: https://jod.li/jod-li-initiatives/distribute-optimised-genetic-algorithm-evolution-with-niche-using-oga-jod-li/

#username and key as defined on https://oga.jod.li
u=toto
k=1234567890123456



#name of the population
pop=bknch5

#used to create population in case it doesn't exist:
mut=0.2
cro=0.2
eli=0.8
ndr=20
psi=1
min=3
max=3000
enc=xdigit
env=csv
nb=10
auto=1
mp=0.5
ck=3
bk=10
ac=0.1
dc=0.1
cc=0.1
sc=0.1
am=1
amp=1
aac=1
adc=1
acc=1
asc=1
acs=1
ae=1
an=0
ap=0


#stop when this fitness is reached
maxfitness=500

#stop after this many chromosome batches
maxchromosomes=200

getfitnessexecutable="./getfitness"

if [[ -x "$getfitnessexecutable" ]]
then
  echo "File '$getfitnessexecutable' is executable. Life is great."
else
  echo "File '$getfitnessexecutable' is not executable or not found. Aborting."
	exit 5
fi

infile="gainputrgb.tmp"
convertCmd=""
#Fitness must be integer
re='^[0-9]+$'

#let read command know the csv separator
export IFS=","


function getBest(){
	#writes out the best chromosome
	echo wget -qO- "https://en.oga.jod.li/?api=gaec&u=$u&k=$k&pop=$pop"
	result=`wget -qO- "https://en.oga.jod.li/?api=gaec&u=$u&k=$k&pop=$pop"`
	echo $result | grep 'Error:' > /dev/null
	if [ $? == 0 ]; then
		echo "failed to get the best chromosome"
		exit 1
	fi
	echo "$result"|head -n 1|while read -r cid chromosome fitness; do
    echo "cid: $cid"
		echo "fitness: $fitness"
    echo ${chromosome}|${getfitnessexecutable} oga-jod-li_best_${u}_${pop}_${fitness}_${cid}
	done
}

sendfitness(){
  fitnessanswer=`wget -qO- "https://en.oga.jod.li/?api=scf&u=$u&k=$k&id=$1&fitness=$2"`
  echo $fitnessanswer | grep 'Error:' > /dev/null
  if [ $? == 0 ]; then
    echo $fitnessanswer | grep 'chromosome fitness too low, deleted' > /dev/null
    if [ $? == 0 ]; then
      echo "warning: chromosome fitness too low, deleted"
    else
      echo "unknown error while sending chromosome fitness, continuing: $fitnessanswer"
      exit 4
    fi
  else
    echo "chromosome $1 fitness sent"
  fi
}

if [ "$#" -gt 0 ];then
	if [ $1 == "best" ];then
		echo "getting best"
		getBest
		exit 0
	fi
fi

#make sure the population is accessible
result=`wget -qO- "https://en.oga.jod.li/?api=gaec&u=$u&k=$k&pop=$pop"`
echo $result | grep 'Error:' > /dev/null
if [ $? == 0 ]; then
	echo $result | grep "Error: population name doesn't exist on get evaluated" > /dev/null
	if [ $? == 0 ]; then
		echo "population does not exist, creating it"
		#creating the population
		result=`wget -qO- "https://en.oga.jod.li/?api=pop&u=$u&k=$k&pop=$pop&mut=$mut&cro=$cro&eli=$eli&ndr=$ndr&psi=$psi&min=$min&max=$max&enc=$enc&env=$env&nb=$nb&auto=$auto&mp=$mp&ck=$ck&bk=$bk&ac=$ac&dc=$dc&cc=$cc&sc=$sc&am=$am&amp=$amp&aac=$aac&adc=$adc&acc=$acc&asc=$asc&acs=$acs&ae=$ae&an=$an&ap=$ap"`
		echo $result | grep 'Error:' > /dev/null
		if [ $? == 0 ]; then
			echo "error creating population, aborting: $result"
			exit 1
		else
			echo "population created"
		fi
	else
		echo "unknown error, aborting: $result"
		exit 2
	fi
fi

error=0
topfitness=0
firstexecution=1
errorfile="gaerrorfile.tmp"
topfitnessfile="gatopfitness.tmp"
nbchromosomesfile="ganbchromosomes.tmp"
echo "0">$errorfile
echo "0">$topfitnessfile
echo "0">$nbchromosomesfile

#loop until maxfitness is reached
while [ "$topfitness" -lt "$maxfitness" ]; do
  nbchromosomes=`cat $nbchromosomesfile`
  if [ $nbchromosomes -ge $maxchromosomes ]; then
    echo "reached maxchromosomes: $maxchromosomes"
    exit 0
  fi
  nbchromosomes=$(( $nbchromosomes + 1 ));echo "batch number: $nbchromosomes"
  echo "$nbchromosomes">$nbchromosomesfile
	#get the chromosomes to evaluate
	result=`wget -qO- "https://en.oga.jod.li/?api=gc&u=$u&k=$k&pop=$pop"`
	echo $result | grep 'Error:' > /dev/null
	if [ $? == 0 ]; then
		echo "unknown error while getting chromosomes to evaluate, aborting: $result"
		echo "3">$errorfile
		exit 3
	fi
	#parse through the chromosomes to evaluate
	echo "$result"|while read -r cid chromosome; do
		#evaluate chromosome
	  echo "cid: $cid"
		fitness=`echo "$chromosome"|$getfitnessexecutable`
		if ! [[ $fitness =~ $re ]] ; then
			echo "Fitness is not an integer: $fitness . Not sending to server." >&2
			exit 6
		fi
		echo "fitness: $fitness"
		if [ "$topfitness" -lt "$fitness" ]; then
			topfitness=$fitness
			echo "$topfitness">$topfitnessfile
			echo "topfitness: $topfitness"
		fi
    if [ $firstexecution -lt 0 ]
    then
      #wait for the previous fork before starting another one
      wait
    else
      firstexecution=0
    fi
		#send fitness to server
    sendfitness $cid $fitness
	done
	error=`cat $errorfile`
	if [ "$error" -gt 0 ]; then
		exit $error
	fi
	topfitness=`cat $topfitnessfile`
	if [ "$topfitness" -ge "$maxfitness" ]; then
		echo "max fitness reached, ending"
		exit 0
	fi
done
error=`cat $errorfile`
rm $errorfile
rm $topfitnessfile
exit $error

#wait for any remaining fork
#wait
